/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>

#include <ox/claw/write.hpp>

#include "gfx.hpp"

namespace nostalgia::core {

struct BaseConverter {
	virtual ~BaseConverter() noexcept = default;

	virtual ox::String srcTypeName() noexcept = 0;

	virtual int srcTypeVersion() noexcept = 0;

	virtual bool srcMatches(const ox::String &srcTypeName, int srcTypeVersion) noexcept = 0;

	virtual bool dstMatches(const ox::String &dstTypeName, int dstTypeVersion) noexcept = 0;

	virtual bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept = 0;

	virtual ox::Error convertRaw(const ox::Buffer &pV1Buff, void *pV2) noexcept = 0;

	virtual ox::Result<ox::Buffer> convertBuffToBuff(const ox::Buffer &srcBuff, ox::ClawFormat dstFmt = ox::ClawFormat::Metal) noexcept = 0;
};

template<typename SrcType, typename DstType>
struct Converter: public BaseConverter{
	ox::String srcTypeName() noexcept final {
		return SrcType::TypeName;
	}

	int srcTypeVersion() noexcept final {
		return SrcType::TypeVersion;
	}

	bool srcMatches(const ox::String &srcTypeName, int srcTypeVersion) noexcept final {
		return srcTypeName == DstType::TypeName
		    && srcTypeVersion == DstType::TypeVersion;
	}

	bool dstMatches(const ox::String &dstTypeName, int dstTypeVersion) noexcept final {
		return dstTypeName == DstType::TypeName
		    && dstTypeVersion == DstType::TypeVersion;
	}

	bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept final {
		return srcTypeName == SrcType::TypeName
		    && dstTypeName == DstType::TypeName
		    && srcTypeVersion == SrcType::TypeVersion
		    && dstTypeVersion == DstType::TypeVersion;
	}

	virtual ox::Error convert(SrcType*, DstType*) noexcept {
		return OxError(1);
	}

	ox::Error convertRaw(const ox::Buffer &pV1Buff, void *pV2) noexcept override {
		oxRequireM(src, ox::readClaw<SrcType>(pV1Buff));
		auto dst = static_cast<DstType*>(pV2);
		return convert(&src, dst);
	}

	ox::Result<ox::Buffer> convertBuffToBuff(const ox::Buffer &srcBuff, ox::ClawFormat dstFmt) noexcept final {
		DstType dst;
		oxReturnError(convertRaw(srcBuff, &dst));
		oxRequireM(out, ox::writeClaw(&dst, dstFmt));
		return out;
	}
};

ox::Result<ox::Buffer> convert(const ox::Buffer &srcBuffer, const ox::String &dstTypeName, int dstTypeVersion, ox::ClawFormat fmt = ox::ClawFormat::Metal) noexcept;

template<typename T>
ox::Result<ox::Buffer> convert(const ox::Buffer &srcBuffer, ox::ClawFormat fmt = ox::ClawFormat::Metal) noexcept {
	return convert(srcBuffer, T::TypeName, T::TypeVersion, fmt);
}

template<typename T>
ox::Error convert(const ox::Buffer &buff, T *outObj) noexcept {
	oxRequireM(outBuff, convert<T>(buff));
	return ox::readClaw<T>(outBuff).moveTo(outObj);
}

}
